"""
Helper functions, primarily for property reports, which are channel reports.
"""

from functools import reduce
import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from emod_api.channelreports.channels import ChannelReport

__all__ = ["property_report_to_csv", "read_json_file", "get_report_channels", "_validate_property_report_channels", "_validate_property_report_ips", "accumulate_trace_data", "__get_trace_name", "save_to_csv", "plot_trace", "__index_for", "__title_for"]


def property_report_to_csv(source_file: Union[str, Path], channels: Optional[List[str]], rollup: Optional[List[str]], csv_file: Union[str, Path], normalize: bool=True, transpose: bool=False) -> None:

    """
    Write a property report to a CSV formatted file.
    
    Optionally selected a subset of available channels.
    Optionally "rolling-up" IP:value sub-channels into a "parent" IP.

    Args:
        source_file: filename of property report
        channels: list of channels to output, None results in writing _all_ channels to output
        rollup: list of IPs into which to roll-up remaining IPs, None indicates no rollup, [] indicates _all_ rolled up
        csv_file: filename of CSV formatted result
        normalize: include a row of normalization data (Statistical Population)
        transpose: write channels as columns rather than rows
    """

    json_data = read_json_file(Path(source_file))
    pool_data = get_report_channels(json_data)

    if channels is None:
        channels = sorted(set([key.split(":")[0] for key in sorted(pool_data.keys())]))
    elif isinstance(channels, str):
        channels = [channels]

    if isinstance(rollup, str):
        rollup = [rollup]

    _validate_property_report_channels(channels, pool_data)
    _validate_property_report_ips(rollup, pool_data)

    (trace_values, norm_values) = accumulate_trace_data(channels, False, rollup, "Statistical Population", pool_data)

    save_to_csv(trace_values, norm_values, csv_file, normalize, transpose)

    return


def read_json_file(filename: Union[str, Path]) -> Dict:

    with Path(filename).open("r") as file:
        json_data = json.load(file)

    return json_data


def get_report_channels(json_data: Dict) -> Dict:

    try:
        channel_data = json_data['Channels']
    except KeyError as e:
        raise KeyError("Didn't find 'Channels' in JSON data.")

    return channel_data


def _validate_property_report_channels(channels, pool_data) -> None:

    if channels:
        keys = set(map(lambda name: name.split(":", 1)[0], pool_data.keys()))
        not_found = [name for name in channels if not name in keys]
        if not_found:
            raise ValueError(f"Specified channel(s) - {not_found} - is/are not valid channel names.")

    return


def _validate_property_report_ips(rollup, pool_data) -> None:

    if rollup:
        first = next(iter(pool_data.keys()))
        ip_string = first.split(":", 1)[1]
        ips = [kvp.split(":")[0] for kvp in ip_string.split(",")]
        not_found = [ip for ip in rollup if not ip in ips]
        if not_found:
            raise ValueError(f"Specified rollup IP(s) - {not_found} - is/are not valid IP names.")

    return


def accumulate_trace_data(channels: List[str], verbose: bool, rollup: List[str], normalize_by: str, poolData: Dict) -> Tuple[Dict[str, np.ndarray], np.ndarray]:

    """
    Extract selected channel(s) from property report data.

    Aggregate on rollup IP(s), if provided, otherwise on channel per unique
    IP:value pair (e.g., "QualityOfCare:High"), per main channel (e.g., "Infected").

    Args:
        channels:       names of channels to plot
        verbose:        output some "debugging"/progress information if true
        rollup:         IP(s) under which to roll-up other IP:value pairs
        normalize_by:   channel for normalization
        poolData:       data for channels keyed on channel name

    Returns:
        tuple of dictionary of aggregated data, keyed on channel name, and of Numpy array of normalization values
    """

    traceValues = {}
    normValues  = {}
    pool_keys = sorted(poolData.keys())

    name_ip_pairs = map(lambda key: tuple(key.split(":", 1)), pool_keys)
    name_ip_pairs_to_process = filter(lambda p: p[0] in channels, name_ip_pairs)
    for (channelTitle, keyValuePairs) in name_ip_pairs_to_process:

        print(f"Processing channel '{channelTitle}:{keyValuePairs}'") if verbose else None

        keyValuePairs = keyValuePairs.split(',')
        traceName = __get_trace_name(channelTitle, keyValuePairs, rollup)
        traceData = np.array(poolData[ f"{channelTitle}:{','.join(keyValuePairs)}" ][ 'Data' ], dtype=np.float32)

        if traceName not in traceValues:
            print(f"New trace: '{traceName}'") if verbose else None
            traceValues[traceName] = traceData
        else:
            print(f"Add to trace: '{traceName}'") if verbose else None
            traceValues[traceName] += traceData

    if normalize_by:
        norm_channels = [channel for channel in poolData.keys() if channel.startswith(f"{normalize_by}:")]
        norm_data = [poolData[channel]["Data"] for channel in norm_channels]
        normValues = reduce(lambda x, y: x + np.array(y, dtype=np.float32), norm_data, 0)
    else:
        normValues = 1

    return (traceValues, normValues)


def __get_trace_name(channelTitle: str, keyValuePairs: List[str], rollup: List[str]) -> str:

    """
    Return "canonical" trace name for a given channel, IP:value list, and rollup list.

    Example:
        title = "Infected"
        keyValuePairs = ["Age_Bin:Age_Bin_Property_From_0_To_20","QualityOfCare:High","QualityOfCare1:High","QualityOfCare2:High"]

        rollup = None
        return "Infected:Age_Bin:Age_Bin_Property_From_0_To_20,QualityOfCare:High,QualityOfCare1:High,QualityOfCare2:High"

        rollup = ["Age_Bin"]
        return = "Infected:Age_Bin:Age_Bin_Property_From_0_To_20"

        rollup = ["Age_Bin", "QualityOfCare"]
        return = "Infected:Age_Bin:Age_Bin_Property_From_0_To_20,QualityOfCare:High"

        rollup = []
        return = "Infected"
    """

    # trace name will have channel title and any property:value pairs
    # which aren't being grouped

    traceName = channelTitle + ':'

    if rollup is None:
        traceName = f"{channelTitle}:{','.join(keyValuePairs)}"
    else:
        if len(rollup) > 0:
            kvps = filter(lambda pair: pair.split(":")[0] in rollup, keyValuePairs)
            traceName = f"{channelTitle}:{','.join(kvps)}"
        else:
            traceName = channelTitle

    return traceName


def save_to_csv(traceValues: Dict[str, np.ndarray], normValues: Union[int, np.ndarray], filename: Union[str, Path], normalize: bool=True, transpose: bool=False) -> None:

    """
    Save property report to CSV. Uses underlying ChannelReport.to_csv() function.

    Args:
        traceValues: full set of available channels, keyed on channel name
        normValues:  normalization values for channels
        filename:    destination file for CSV data
        normalize:   whether or not to include normalization data (in a final, "normalize" row of the CSV)
        transpose: write channels as columns rather than rows
    """

    cr = ChannelReport()

    for channel, data in traceValues.items():
        cr.channels[channel] = data

    if normalize:
        cr.channels["normalize"] = list(normValues)     # list from np.ndarray

    cr.to_csv(Path(filename), transpose=transpose)   # by default, use _all_ the channels we just added

    return


def plot_traces(traceValues: Dict[str, np.ndarray], normValues: Union[int, np.ndarray], overlay: bool, channels: List[str], normalize: bool, title: str, legend: bool, by: str) -> plt.Figure:

    """
    Plot channel data. One subplot per channel unless overlaying all variations of rolled-up IP(s) is requested.

    Args:
        traceValues: channel data, keyed on channel name
        normValues:  normalization data for channels
        overlay:     whether or not to overlay all variations of a given channel on one subplot
        channels:    selection of channel names to plot
        normalize:   whether or not to normalize the channel data with the normalization data
        title:       plot title
        legend:      whether or not to include a legend on plots
        by:          name of channel being used for normalization

    Returns:
        plt.Figure
    """

    if len(traceValues) == 0:
        print("Didn't find requested channel(s) in property report.")
        return

    if not overlay:
        plotCount = len(traceValues)
    else:
        plotCount = len(channels)

    if normalize:
        plotCount *= 2

    figure = plt.figure(title, figsize=(16,9), dpi=300)
    traceKeys = sorted(traceValues.keys())

    # plotting here
    for traceName in traceKeys:
        plotIndex = __index_for(traceName, channels, traceKeys, normalize, overlay)
        plt.subplot(plotCount, 1, plotIndex)
        plt.plot(traceValues[traceName], label=traceName)
        if normalize:
            plt.subplot(plotCount, 1, plotIndex+1)
            plt.ylim((0.0, 1.0))    # yes, this takes a tuple
            plt.plot(traceValues[traceName]/normValues, label=traceName)

    # make it pretty
    _ = plt.subplot(plotCount, 1, 1)
    for traceName in traceKeys:
        plotIndex = __index_for(traceName, channels, traceKeys, normalize, overlay)
        plotTitle = __title_for(traceName, channels, overlay)
        plt.subplot(plotCount, 1, plotIndex)
        plt.title(plotTitle)
        plt.legend() if legend else None
        if normalize:
            plt.subplot(plotCount, 1, plotIndex+1)
            plt.title(f"{plotTitle} normalized by {by}")
            plt.legend() if legend else None

    plt.tight_layout()

    return figure


def __index_for(traceName: str, channels: List[str], traceKeys: List[str], normalize: bool, overlay: bool) -> int:

    if overlay:
        # all pools of the same channel overlaid
        index = 0
        for channel in channels:
            if channel in traceName:
                break
            index += 1
    else:
        # each trace separate
        index = traceKeys.index(traceName)

    # if we're normalizing, there's a normalized trace per regular trace
    if normalize:
        index *= 2

    # matplotlib is 1-based (like MATLAB)
    return index+1


def __title_for(traceName: str, channels: List[str], overlay: bool):

    # use channel name
    if overlay:
        for channel in channels:
            if channel in traceName:
                title = channel
                break
    else:
        title = traceName

    return title
